#****************************************************************************
#**
#**  File     :  /lua/unit.lua
#**  Author(s):  John Comes, David Tomandl, Gordon Duclos
#**
#**  Summary  : The Unit lua module
#**
#**  Copyright  2005 Gas Powered Games, Inc.  All rights reserved.
#****************************************************************************

local oldUnit=Unit
Unit = Class(oldUnit) {
    
	OnCreate = function(self)
    	oldUnit.OnCreate(self)
    	local bp = self:GetBlueprint()
    	self.VeteranLevel = 0
    	self.xp = 0
    	self.XPnextLevel = bp.Economy.XPperLevel
    	self.LevelProgress = 0
    	#self.Sync.VeteranLevel = self.VeteranLevel
        self.Sync.LevelProgress = self.LevelProgress
        if bp.Economy.BuildRate then
        	self.Sync.BuildRate = bp.Economy.BuildRate
        end
        self.Sync.RegenRate = bp.Defense.RegenRate
    end, 

	OnStopBeingBuilt = function(self, builder, layer)
		if self:GetBlueprint().Economy.xpTimeStep then
        	self:ForkThread(self.XPOverTime)
        end
        oldUnit.OnStopBeingBuilt(self, builder, layer)
	end,
    
	XPOverTime = function(self)
	    local waittime = self:GetBlueprint().Economy.xpTimeStep / 100
		WaitSeconds(waittime)
	    while not self:IsDead() do 
	        #self:AddLevels(0.01)
	    	self:AddXP(self.XPnextLevel/100)	
			WaitSeconds(waittime)
	    end
    end,
    
    startBuildXPThread = function(self)
        local levelPerSecond = self:GetBlueprint().Economy.BuildXPLevelpSecond
		if not levelPerSecond then return end
        WaitSeconds(1)
	    while not self:IsDead() do 
	    	self:AddXP(self.XPnextLevel*levelPerSecond)	
			WaitSeconds(1)
	    end
    end,
     
    OnStartBuild = function(self, unitBeingBuilt, order)
   	    #LOG('unit:OnStartBuild')
   	    self.BuildXPThread = ForkThread(self.startBuildXPThread, self)
        oldUnit.OnStartBuild(self, unitBeingBuilt, order)    
    end,

    OnStopBuild = function(self, unitBeingBuilt)
        #LOG('unit:OnStopBuild')
        KillThread(self.BuildXPThread)
        oldUnit.OnStopBuild(self, unitBeingBuilt)
    end,

    OnFailedToBuild = function(self)
    	#LOG('unit:OnFailedToBuild')
    	KillThread(self.BuildXPThread)
    	oldUnit.OnFailedToBuild(self)
    end,
    
    
    ##########################################################################################
    ## VETERANCY
    ##########################################################################################
    
    #This function should be used for kills made through the script, since kills through the engine (projectiles etc...) are already counted.
    AddKills = function(self, numKills)
        #Add the kills, then check veterancy junk.
        local unitKills = self:GetStat('KILLS', 0).Value + numKills
        self:SetStat('KILLS', unitKills)
        
        local vet = self:GetBlueprint().Veteran or Game.VeteranDefault
        
        local vetLevels = table.getsize(vet)
        if self.VeteranLevel == vetLevels then
            return
        end

        local nextLvl = self.VeteranLevel + 1
        local nextKills = vet[('Level' .. nextLvl)]
        
        #Since we could potentially be gaining a lot of kills here, check if we gained more than one level
        while unitKills >= nextKills and self.VeteranLevel ~= vetLevels do
            self:SetVeteranLevel(nextLvl)
            
            nextLvl = self.VeteranLevel + 1
            nextKills = vet[('Level' .. nextLvl)]
        end 
    end,

    
    
    # use this to go through the AddKills function rather than directly setting veterancy
    SetVeterancy = function(self, veteranLevel)
        veteranLevel = veteranLevel or 0
        if veteranLevel == 0 then
            return
        end
        local bp = self:GetBlueprint()
        if bp.Veteran['Level1'] then
			##LOG('unit:AddKills')
            self:AddKills(bp.Veteran['Level1']*veteranLevel)
        elseif import('/lua/game.lua').VeteranDefault['Level'..veteranLevel] then
            ##LOG('VeteranDefault.unit:AddKills')
        	self:AddKills(import('/lua/game.lua').VeteranDefault['Level'..veteranLevel])
        else
            error('Invalid veteran level - ' .. veteranLevel)
        end 
    end, 

    #Check to see if we should veteran up.
    #CheckVeteranLevel = function(self)
    #    local bp = self:GetBlueprint().Veteran
    #    #There is no veteran block in the bp, return
    #    if not bp then
    #        bp = Game.VeteranDefault
    #    end
    #    #LOG('*DEBUG: Vet Table = ', repr(bp))
    #    #We add 1 because we get this before the stat gets updated
    #    local unitKills = self:GetStat('KILLS', 0).Value + 1
    #    #LOG('*DEBUG: Veteran Kills = ', repr(unitKills))
    #    #We are already at the highest veteran level, return
    #    #if self.VeteranLevel == table.getsize(bp) then
    #    #    return
    #    #end

    #    local nextLvl = self.VeteranLevel + 1
    #    local nextKills = bp[('Level1')]*nextLvl
    #    if unitKills >= nextKills then
    #    	self:SetVeteranLevel(nextLvl)
    #    end
    #end,

    #Check to see if we should veteran up.
    CheckVeteranLevel = function(self)
        local bp = self:GetBlueprint()
        #LOG('xp:' .. self.xp.. ' level at:' .. self.XPperLevel)
		while self.xp >= self.XPnextLevel do
            self.xp = self.xp - self.XPnextLevel
			self:SetVeteranLevel(self.VeteranLevel + 1)
			self.XPnextLevel = bp.Economy.XPperLevel * (1+ 0.05*self.VeteranLevel)
        end
        self.LevelProgress = self.xp / self.XPnextLevel + self.VeteranLevel
        self.Sync.LevelProgress = self.LevelProgress
    end,
    
    AddLevels = function(self, levels)
		local bp = self:GetBlueprint()
		local curlevel = self.VeteranLevel
		local percent = self.LevelProgress - curlevel
		
		local xpAdd = 0
		if levels >= (1-percent) then
			xpAdd = self.XPnextLevel * (1-percent)
			levels=levels-(1-percent)
		else
			xpAdd =self.XPnextLevel * levels
			levels=0
		end
			
		while levels > 1 do
			levels=levels-1
			curlevel = curlevel +1
			xpAdd=xpAdd + bp.Economy.XPperLevel * (1+ 0.05*curlevel)
		end
		xpAdd=xpAdd + bp.Economy.XPperLevel * (1+ 0.05*(curlevel+1)) * levels
		self:AddXP(xpAdd)
    end,
    
    #Set the veteran level to the level specified
    SetVeteranLevel = function(self, level)
        #LOG(' ')
        #LOG('*DEBUG: '.. self:GetBlueprint().Description .. ' VETERAN UP! LEVEL ', repr(level))
        local old = self.VeteranLevel
        self.VeteranLevel = level

        # Apply default veterancy buffs
        local buffTypes = { 'Regen', 'Health', 'Damage','DamageArea','Range','Speed','Vision','OmniRadius','Radar','BuildRate','ResourceProduction','RateOfFire','Shield'}
        for k,bType in buffTypes do
            Buff.ApplyBuff( self, 'Veterancy' .. bType)
            #Buff.ApplyBuff( self, 'Veterancy' .. bType .. level )
        end
        
        # Get any overriding buffs if they exist
        local bp = self:GetBlueprint().Buffs
        #Check for unit buffs
        
   		if bp then
        	for bLevel,bData in bp do
        		if (bLevel == 'Any' or bLevel == 'Level'..level) then
        			for bType,bValues in bData do
	        			local buffName = self:CreateUnitBuff(bLevel,bType,bValues)
	        			if buffName then
	        				Buff.ApplyBuff( self, buffName )
	                    end
        			end
        		end
        	end   		
        end

        self:GetAIBrain():OnBrainUnitVeterancyLevel(self, level)
        self:DoUnitCallbacks('OnVeteran')
    end,
    
    #function to generate the new veterancy buffs
    CreateUnitBuff = function(self, levelName, buffType, buffValues)
    	
    	# Generate a buff based on the unitId
        local buffName = self:GetUnitId() .. levelName .. buffType
        local buffMinLevel = nil
        local buffMaxLevel = nil
        if buffValues.MinLevel then buffMinLevel = buffValues.MinLevel end
        if buffValues.MaxLevel then buffMaxLevel = buffValues.MaxLevel end
        
        
        # Create the buff if needed
        if not Buffs[buffName] then
        	#LOG(buffName .. ': '..buffMinLevel.. ' - '..buffMaxLevel)
            BuffBlueprint {
             	MinLevel = buffMinLevel,
            	MaxLevel = buffMaxLevel,
                Name = buffName,
                DisplayName = buffName,
                BuffType = buffType,
                Stacks = buffValues.Stacks,
                #self.BuffTypes[buffType].BuffStacks,
                Duration = buffValues.Duration,
                Affects = buffValues.Affects,
            }
        end
        
        # Return the buffname so the buff can be applied to the unit
        return buffName
    end,
   
    #Allowing buff bonus anhd adj bonus at same time!!
  	UpdateProductionValues = function(self)
        local bpEcon = self:GetBlueprint().Economy
        if not bpEcon then return end
        self:SetProductionPerSecondEnergy( (self.EnergyProdMod or bpEcon.ProductionPerSecondEnergy or 0)* (self.EnergyProdAdjMod or 1) )
        self:SetProductionPerSecondMass( (self.MassProdMod or bpEcon.ProductionPerSecondMass or 0) * (self.MassProdAdjMod or 1) )
    end,
    
    AddXP = function(self,amount)
    	if not self.XPnextLevel then return end
    	self.xp = self.xp + amount
    	self:CheckVeteranLevel()
    end,
    
    DoTakeDamage = function(self, instigator, amount, vector, damageType)
    	if instigator and IsUnit(instigator) and not instigator:IsDead() then
	    	local preAdjHealth = self:GetHealth()
	    	local bp = self:GetBlueprint()
	    	if amount>=preAdjHealth then
	    		instigator:AddXP(preAdjHealth*bp.Economy.xpPerHp)
	    	else
	    		instigator:AddXP(amount*bp.Economy.xpPerHp)
	    		self:AddXP(amount*bp.Economy.xpPerHp)
	    	end
    	end
    	oldUnit.DoTakeDamage(self, instigator, amount, vector, damageType)
        
    end,
    
 	GetShield = function(self)
        return self.MyShield or nil
    end,
}

